# -*- coding: utf-8 -*-

from types import GetSetDescriptorType, MemberDescriptorType
from typing import TYPE_CHECKING, cast

from six import iteritems, with_metaclass

if TYPE_CHECKING:
    from typing import Any, Dict, Tuple, Type

    SlotsTuple = Tuple[str, ...]
    MembersDict = Dict[str, Dict[Type, MemberDescriptorType]]
    StateDict = Dict[str, Dict[Type, Any]]


def privatize_name(cls_name, name):
    # type: (str, str) -> str
    """Privatize an attribute name if necessary."""
    if name.startswith("__") and not name.endswith("__"):
        return "_{}{}".format(cls_name.lstrip("_"), name)
    return name


def update_members(members, members_update):
    # type: (MembersDict, MembersDict) -> None
    """Update members dictionary in place."""
    for name, bases in iteritems(members_update):
        members.setdefault(name, {})
        for base, member in iteritems(bases):
            members[name][base] = member


def scrape_members(base):
    # type: (Type) -> MembersDict
    """Scrape base for members (shallow)."""
    if base is object:
        return {}
    base_members = {}  # type: MembersDict
    for name in base.__slots__:
        private_name = privatize_name(base.__name__, cast(str, name))
        try:
            member = base.__dict__[private_name]
        except KeyError:
            continue
        if isinstance(member, MemberDescriptorType):
            if member.__objclass__ is base and member.__name__ == private_name:
                base_members.setdefault(private_name, {})
                base_members[private_name][base] = member
    return base_members


def scrape_all_members(base):
    # type: (Type) -> MembersDict
    """Scrape base for all members (deep)."""
    all_members = {}  # type: MembersDict
    for base_base in reversed(base.__mro__):
        if base_base is base:
            update_members(all_members, scrape_members(base))
        else:
            update_members(all_members, scrape_all_members(base_base))
    return all_members


def get_state(obj):
    # type: (Slotted) -> StateDict
    """Get state of a slotted object for pickling purposes."""
    state = {}  # type: StateDict
    for name, members in iteritems(type(obj).__members__):
        for base, member in iteritems(members):
            try:
                value = member.__get__(obj, base)
            except AttributeError:
                continue
            state.setdefault(name, {})
            state[name][base] = value
    return state


def set_state(obj, state):
    # type: (Slotted, StateDict) -> None
    """Set state of a slotted object for unpickling purposes."""
    for name, bases in iteritems(state):
        for base, value in iteritems(bases):
            member = base.__dict__[name]
            if isinstance(member, MemberDescriptorType):
                if member.__objclass__ is base and member.__name__ == name:
                    getattr(member, "__set__")(obj, value)


class SlottedMeta(type):
    """Enforces usage of '__slots__'."""

    @staticmethod
    def __new__(
        mcs,  # type: Type[SlottedMeta]
        name,  # type: str
        bases,  # type: Tuple[Type, ...]
        dct,  # type: Dict[str, Any]
    ):
        # type: (...) -> SlottedMeta
        """Make slotted class."""
        for base in bases:
            if isinstance(base.__dict__.get("__dict__"), GetSetDescriptorType):
                error = "base '{}' does not enforce '__slots__'".format(base.__name__)
                raise TypeError(error)
        dct = dict(dct)
        dct["__slots__"] = tuple(dct.get("__slots__", ()))
        return cast(SlottedMeta, super(SlottedMeta, mcs).__new__(mcs, name, bases, dct))

    @property
    def __members__(cls):
        # type: () -> MembersDict
        members = {}  # type: MembersDict
        for base in reversed(cls.__mro__):
            if base is cls:
                base_members = scrape_members(cls)
            elif isinstance(base, SlottedMeta):
                base_members = base.__members__
            else:
                base_members = scrape_all_members(base)
            update_members(members, base_members)
        return members


class Slotted(with_metaclass(SlottedMeta, object)):
    """Enforces usage of '__slots__' and implements pickling methods based on state."""

    __slots__ = ()  # type: SlotsTuple

    def __getstate__(self):
        # type: () -> StateDict
        """Get state for pickling."""
        return get_state(self)

    def __setstate__(self, state):
        # type: (StateDict) -> None
        """Set state for unpickling."""
        set_state(self, state)



# -*- coding: utf-8 -*-

from abc import ABCMeta
from types import MemberDescriptorType
from typing import TYPE_CHECKING, cast

from six import iteritems, with_metaclass
from six.moves import collections_abc

from ._bases import Slotted, SlottedMeta, privatize_name

if TYPE_CHECKING:
    from typing import Any, Dict, List, Set, Tuple, Type, Union

__all__ = [
    "SlottedABCMeta",
    "SlottedABC",
    "SlottedCallable",
    "SlottedContainer",
    "SlottedHashable",
    "SlottedItemsView",
    "SlottedIterable",
    "SlottedIterator",
    "SlottedKeysView",
    "SlottedMapping",
    "SlottedMappingView",
    "SlottedMutableMapping",
    "SlottedMutableSequence",
    "SlottedMutableSet",
    "SlottedSequence",
    "SlottedSet",
    "SlottedSized",
    "SlottedValuesView",
]

_ABC_ALL = [
    "ABCMeta",
    "Callable",
    "Container",
    "Hashable",
    "ItemsView",
    "Iterable",
    "Iterator",
    "KeysView",
    "Mapping",
    "MappingView",
    "MutableMapping",
    "MutableSequence",
    "MutableSet",
    "Sequence",
    "Set",
    "Sized",
    "ValuesView",
]

SlottedCallable = collections_abc.Callable
SlottedContainer = collections_abc.Container
SlottedHashable = collections_abc.Hashable
SlottedItemsView = collections_abc.ItemsView
SlottedIterable = collections_abc.Iterable
SlottedIterator = collections_abc.Iterator
SlottedKeysView = collections_abc.KeysView
SlottedMapping = collections_abc.Mapping
SlottedMappingView = collections_abc.MappingView
SlottedMutableMapping = collections_abc.MutableMapping
SlottedMutableSequence = collections_abc.MutableSequence
SlottedMutableSet = collections_abc.MutableSet
SlottedSequence = collections_abc.Sequence
SlottedSet = collections_abc.Set
SlottedSized = collections_abc.Sized
SlottedValuesView = collections_abc.ValuesView


class SlottedABCMeta(SlottedMeta, ABCMeta):
    """Metaclass for 'SlottedABC'. Combines 'SlottedMeta' and 'ABCMeta'."""


class SlottedABC(with_metaclass(SlottedABCMeta, Slotted)):
    """Slotted 'ABC' base class."""


_CLASSES = set()  # type: Set[Type]
for cls_name in _ABC_ALL:
    try:
        cls = getattr(collections_abc, cls_name)
    except AttributeError:
        continue
    if not isinstance(cls, ABCMeta) or not issubclass(cls, object):
        continue
    _CLASSES.add(cls)

_CACHE = {
    object: SlottedABC,
    ABCMeta: SlottedABCMeta,
}  # type: Dict[Type[Any], Union[SlottedABCMeta, Type[SlottedABCMeta]]]


def extract_dict(base):
    # type: (Type) -> Tuple[Dict[str, Any], Dict[str, Any]]
    slots = set(
        privatize_name(base.__name__, slot) for slot in getattr(base, "__slots__", ())
    )
    base_dict = {}
    overrides = {}
    for name, value in iteritems(base.__dict__):
        if name in slots:
            if isinstance(value, MemberDescriptorType):
                if value.__objclass__ is base and value.__name__ == name:
                    continue
            overrides[name] = value
        else:
            base_dict[name] = value
    return base_dict, overrides


def convert_meta(source):
    # type: (Type) -> Type
    """Convert an ABCMeta-based metaclass to a SlottedABCMeta-based metaclass."""
    try:
        return _CACHE[source]
    except KeyError:
        pass

    target_bases_list = []  # type: List[Union[SlottedABCMeta, Type[SlottedABCMeta]]]
    for source_base in source.__bases__:
        target_bases_list.append(cast(SlottedABCMeta, convert_meta(source_base)))

    source_name = source.__name__
    target_name = "Slotted{}".format(source_name)
    target_bases = tuple(target_bases_list)
    target_dct, overrides = extract_dict(source)
    target_dct["__module__"] = __name__

    target = SlottedABCMeta(target_name, target_bases, target_dct)
    for name, value in iteritems(overrides):
        type.__setattr__(cast(type, target), name, value)
    _CACHE[source] = target

    return target


def convert(source):
    # type: (ABCMeta) -> Union[SlottedABCMeta, Type[SlottedABCMeta]]
    """Convert an ABC-based class to an SlottedABC-based class."""
    try:
        return _CACHE[source]
    except KeyError:
        pass

    meta = convert_meta(type(source))

    target_bases_list = []  # type: List[Union[SlottedABCMeta, Type[SlottedABCMeta]]]
    for source_base in source.__bases__:
        target_bases_list.append(convert(cast(ABCMeta, source_base)))

    source_name = source.__name__
    target_name = "Slotted{}".format(source_name)
    target_bases = tuple(target_bases_list)
    target_dct, overrides = extract_dict(source)
    target_dct.pop("__dict__", None)
    target_dct["__module__"] = __name__

    target = meta(target_name, target_bases, target_dct)
    for name, value in iteritems(overrides):
        type.__setattr__(target, name, value)
    _CACHE[source] = target

    if target.__dict__.get("__dict__") is not None:
        error = "class '{}' has a '__dict__'".format(target_name)
        raise AssertionError(error)

    source.register(target)
    return target


# Convert all classes
for cls in _CLASSES:
    converted = convert(cast(ABCMeta, cls))
    locals()[converted.__name__] = converted



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="slotted",
    version="1.1.1",
    author="Bruno Nicko",
    author_email="brunonicko@gmail.com",
    description="Enforces usage of '__slots__' for python classes",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/brunonicko/slotted",
    packages=setuptools.find_packages(),
    package_data={"slotted": ["__init__.pyi", "py.typed"]},
    install_requires=["six", "typing; python_version < '3.5'"],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    python_requires=">=2.7",
)



# -*- coding: utf-8 -*-
"""Package stub file."""

from abc import ABC, ABCMeta
from typing import (
    AbstractSet,
    Any,
    Callable,
    Container,
    Dict,
    Hashable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Mapping,
    MappingView,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Sized,
    Type,
    ValuesView,
)

__all__ = [
    "get_state",
    "set_state",
    "SlottedMeta",
    "Slotted",
    "SlottedABCMeta",
    "SlottedABC",
    "SlottedCallable",
    "SlottedContainer",
    "SlottedHashable",
    "SlottedItemsView",
    "SlottedIterable",
    "SlottedIterator",
    "SlottedKeysView",
    "SlottedMapping",
    "SlottedMappingView",
    "SlottedMutableMapping",
    "SlottedMutableSequence",
    "SlottedMutableSet",
    "SlottedSequence",
    "SlottedSet",
    "SlottedSized",
    "SlottedValuesView",
]

def get_state(obj):  # type: (Slotted) -> Dict[str, Dict[Type, Any]]
    pass

def set_state(obj, state):  # type: (Slotted, Dict[str, Dict[Type, Any]]) -> None
    pass

class SlottedMeta(type):
    pass

class Slotted(metaclass=SlottedMeta):
    pass

class SlottedABCMeta(SlottedMeta, ABCMeta):
    pass

class SlottedABC(Slotted, ABC, metaclass=SlottedABCMeta):
    pass

SlottedCallable = Callable

class SlottedContainer(Container, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedHashable(Hashable, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedItemsView(ItemsView, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedIterable(Iterable, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedIterator(Iterator, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedKeysView(KeysView, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedMapping(Mapping, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedMappingView(MappingView, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedMutableMapping(MutableMapping, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedMutableSequence(MutableSequence, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedMutableSet(MutableSet, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedSequence(Sequence, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedSet(AbstractSet, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedSized(Sized, SlottedABC, metaclass=ABCMeta):
    pass

class SlottedValuesView(ValuesView, SlottedABC, metaclass=ABCMeta):
    pass
